﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace Models.Aqo
{
    public class AFPWriter : IDisposable
    {
        public BinaryWriter             Writer                  { get; set; }

        public AFPWriter(string file)
        {
            Writer = new BinaryWriter(File.OpenWrite(file));
        }

        public AFPWriter(Stream s)
        {
            Writer = new BinaryWriter(s);
        }

        public void Dispose()
        {
            if(null != Writer)
            {
                Writer.Dispose();

                Writer  = null;
            }
        }

        public void Write(AFPFile data)
        {
            var w   = Writer;

            SaveHeaders(data);

            if(null != data.VTBF)
            {
                w.Write(Encoding.ASCII.GetBytes(data.VTBF));
                w.Write(data.Unknown1);
                w.Write(Encoding.ASCII.GetBytes(data.AQGF));
                w.Write(data.Unknown2);

                SaveChunks(data);
            }
        }

        private void SaveHeaders(AFPFile data)
        {
            var w   = Writer;

            if(null != data.afp)
            {
                w.Write(Encoding.ASCII.GetBytes(data.afp));
                w.Write(data.Unknown0);
                w.Write(data.Size1);
                w.Write(data.Size2);
                w.Write(data.Size3);
            }

            if(null != data.aqo)
                w.Write(Encoding.ASCII.GetBytes(data.aqo));
        }

        private void SaveChunks(AFPFile data)
        {
            var w   = Writer;

            foreach(var i in data.VTC0)
            {
                w.Write(Encoding.ASCII.GetBytes(i.Type));

                using(var ms= new MemoryStream())
                using(var bw= new BinaryWriter(ms))
                {
                    bw.Write(i.DataObj.Head);
                    bw.Write((ushort)i.DataObj.Params.Count);

                    foreach(var j in i.DataObj.Params)
                        Param.WriteParam(bw, j);

                    ms.Flush();

                    var buf = ms.ToArray();
                    #if false
                    var x   = buf.Length % 16;
                    x       = x == 0 ? 0 : 16 - x;

                    w.Write(buf.Length + x);
                    w.Write(buf);

                    for(int j= 0; j < x; ++j)
                        w.Write((byte)0);
                    #else
                    w.Write(buf.Length + 4);
                    w.Write(Encoding.ASCII.GetBytes(i.SubType));
                    w.Write(buf);
                    #endif
                }
            }

            var size1   = w.BaseStream.Position - 0x40;

          //w.Write(0); // 不要かも
          //w.Write(0);

            while((w.BaseStream.Position % 16) != 0)
            {
                w.Write((byte)0);
                w.Flush();
            }

            var size2   = w.BaseStream.Position;

            w.Flush();
            w.BaseStream.SetLength(w.BaseStream.Position);

            if(null != data.afp)
            {
                w.BaseStream.Position   = 0x30;

                w.Write((int)size1);
                w.Write(0x30);
                w.Write((int)size2);
            }
        }

        public static void Assert(bool b)
        {
            if(!b)
                throw new Exception();
        }
    }
}
