﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace Models.Aqo
{
    public class Param
    {
        public long                     Posision                { get; set; }
        public ushort                   ID                      { get; set; }
        public byte                     Sub1                    { get; set; }
        public byte                     Sub2                    { get; set; }
        public object                   Value                   { get; set; }

        public Param()
        {
        }

        public Param(ushort id)
        {
            ID     = id;
        }

        public Param(ushort id, object value)
        {
            ID      = id;
            Value   = value;
        }

        public static IEnumerable<Param> ParseParams(byte[] data)
        {
            using(var s= new MemoryStream(data))
            using(var r= new BinaryReader(s))
            {
                s.Position  = 2;
                var n       = (int)r.ReadUInt16();

                for(int i= 0; i < n; ++i)
                {
                    var pos = s.Position;
                    var p   = ReadParam(r);

                    yield return p;
                }
            }
        }

        #region Write
        public static void WriteParam(BinaryWriter w, Param p)
        {
            w.Write(p.ID);

            if(p.Value == null)             return;
            if(p.Value is string)           { Write(w, (string)        p.Value);    return; }
            if(p.Value is short)            { Write(w, (short)         p.Value);    return; }
            if(p.Value is int)              { Write(w, (int)           p.Value);    return; }
            if(p.Value is ushort)           { Write(w, (ushort)        p.Value);    return; }
            if(p.Value is uint)             { Write(w, (uint)          p.Value);    return; }
            if(p.Value is float)            { Write(w, (float)         p.Value);    return; }
            if(p.Value is SlimDX.Vector2)   { Write(w, (SlimDX.Vector2)p.Value, p); return; }
            if(p.Value is SlimDX.Vector3)   { Write(w, (SlimDX.Vector3)p.Value, p); return; }
            if(p.Value is SlimDX.Vector4)   { Write(w, (SlimDX.Vector4)p.Value, p); return; }
            if(p.Value is SlimDX.Matrix)    { Write(w, (SlimDX.Matrix) p.Value, p); return; }
            if(p.Value is byte[])           { Write(w, (byte[])        p.Value);    return; }
            if(p.Value is int[])            { Write(w, (int[])         p.Value);    return; }
            if(p.Value is uint[])           { Write(w, (uint[])        p.Value);    return; }
            if(p.Value is ushort[])         { Write(w, (ushort[])      p.Value);    return; }

            throw new NotSupportedException();
        }

        public static void WriteSize(BinaryWriter w, int n)
        {
            if(n <= 256)
            {
                w.Write((byte)8);
                w.Write((byte)(n-1));
            } else
            if(n <= 65536)
            {
                w.Write((byte)16);
                w.Write((ushort)(n-1));
            } else
            {
                w.Write((byte)24);
                w.Write((uint)(n-1));
            }
        }

        public static void Write(BinaryWriter w, string value)
        {
            var data= Encoding.Default.GetBytes(value);

            if(data.Length >= 256)
                throw new FormatException();

            w.Write((byte)data.Length);
            w.Write(data);
        }

        public static void Write(BinaryWriter w, short value)
        {
            w.Write(value);
        }

        public static void Write(BinaryWriter w, int value)
        {
            w.Write(value);
        }

        public static void Write(BinaryWriter w, ushort value)
        {
            w.Write(value);
        }

        public static void Write(BinaryWriter w, uint value)
        {
            w.Write(value);
        }

        public static void Write(BinaryWriter w, float value)
        {
            w.Write(value);
        }

        public static void Write(BinaryWriter w, SlimDX.Vector2 value, Param p)
        {
            w.Write(p.Sub1);
            w.Write(value.X);
            w.Write(value.Y);
        }

        public static void Write(BinaryWriter w, SlimDX.Vector3 value, Param p)
        {
            w.Write(p.Sub1);
            w.Write(value.X);
            w.Write(value.Y);
            w.Write(value.Z);
        }

        public static void Write(BinaryWriter w, SlimDX.Vector4 value, Param p)
        {
            w.Write(p.Sub1);
            w.Write(value.X);
            w.Write(value.Y);
            w.Write(value.Z);
            w.Write(value.W);
        }

        public static void Write(BinaryWriter w, SlimDX.Matrix value, Param p)
        {
            w.Write(p.Sub1);
            w.Write(p.Sub2);
            w.Write(value.M11); w.Write(value.M12); w.Write(value.M13); w.Write(value.M14);
            w.Write(value.M21); w.Write(value.M22); w.Write(value.M23); w.Write(value.M24);
            w.Write(value.M31); w.Write(value.M32); w.Write(value.M33); w.Write(value.M34);
            w.Write(value.M41); w.Write(value.M42); w.Write(value.M43); w.Write(value.M44);
        }

        public static void Write(BinaryWriter w, byte[] value)
        {
            WriteSize(w, value.Length / 4);

            w.Write(value);
        }

        public static void Write(BinaryWriter w, int[] value)
        {
            WriteSize(w, value.Length);

            foreach(var i in value)
                w.Write(i);
        }

        public static void Write(BinaryWriter w, uint[] value)
        {
            WriteSize(w, value.Length);

            foreach(var i in value)
                w.Write(i);
        }

        public static void Write(BinaryWriter w, ushort[] value)
        {
            WriteSize(w, value.Length);

            foreach(var i in value)
                w.Write(i);
        }
        #endregion

        #region Read
        public static Param ReadParam(BinaryReader r)
        {
            var pos = r.BaseStream.Position;
            var id  = r.ReadUInt16();
            var p   = new Param() { ID= id, Posision= pos };

            switch(id & 0xFF00)
            {
            case 0x0000:                            break;
            case 0x0200:    p.ReadString(r);        break;
            case 0x0600:    p.ReadUShort(r);        break;
          //case 0x0800:    p.ReadUInt(r);          break;
            case 0x0800:    p.ReadInt(r);           break;
            case 0x0900:    p.ReadUInt(r);          break;
            case 0x0A00:    p.ReadSingle(r);        break;
            case 0x8600:    p.ReadUshortArray(r);   break;
          //case 0x8800:    p.ReadUIntArray(r);     break;
            case 0x8800:    p.ReadIntArray(r);      break;
            case 0x8900:    p.ReadUIntArray(r);     break;
            case 0xCA00:    p.ReadMatrix(r);        break;
            case 0x4A00:    p.ReadVector(r);        break;
            default:
                throw new Exception();
            }

            return p;
        }

        public static int ReadSize(BinaryReader r)
        {
            var vtype   = r.ReadByte();

            switch(vtype)
            {
            case  8: return r.ReadByte()   + 1;
            case 16: return r.ReadUInt16() + 1;
            case 24: return r.ReadInt32()  + 1;
            default:    throw new Exception();
            }
        }

        public void ReadInt(BinaryReader r)
        {
            var type= ID & 0xFF00;

            if(!(type == 0x0800 || type == 0x0900))
                throw new Exception();

            Value   = r.ReadInt32();
        }

        public void ReadUInt(BinaryReader r)
        {
            var type= ID & 0xFF00;

            if(!(type == 0x0800 || type == 0x0900))
                throw new Exception();

            Value   = r.ReadUInt32();
        }

        public void ReadUShort(BinaryReader r)
        {
            var type= ID & 0xFF00;

            if(!(type == 0x0600))
                throw new Exception();

            Value   = r.ReadUInt16();
        }

        public void ReadSingle(BinaryReader r)
        {
            var type= ID & 0xFF00;

            if(!(type == 0x0A00))
                throw new Exception();

            Value   = r.ReadSingle();
        }

        public void ReadUshortArray(BinaryReader r)
        {
            var type= ID & 0xFF00;

            if(!(type == 0x8600))
                throw new Exception();

            Value = Enumerable.Range(0, ReadSize(r)).Select(j => r.ReadUInt16()).ToArray();
        }

        public void ReadUIntArray(BinaryReader r)
        {
            var type= ID & 0xFF00;

            if(!(type == 0x8800 || type == 0x8900))
                throw new Exception();

            if(ID == 0x89BA)
                    Value = r.ReadBytes(4*ReadSize(r));
            else    Value = Enumerable.Range(0, ReadSize(r)).Select(j => r.ReadUInt32()).ToArray();
        }

        public void ReadIntArray(BinaryReader r)
        {
            var type= ID & 0xFF00;

            if(!(type == 0x8800 || type == 0x8900))
                throw new Exception();

            if(ID == 0x89BA)
                    Value = r.ReadBytes(4*ReadSize(r));
            else    Value = Enumerable.Range(0, ReadSize(r)).Select(j => r.ReadInt32()).ToArray();
        }
        
        public void ReadString(BinaryReader r)
        {
            var type= ID & 0xFF00;

            if(!(type == 0x0200))
                throw new Exception();

            Value = Encoding.Default.GetString(r.ReadBytes(r.ReadByte()));
        }

        public void ReadMatrix(BinaryReader r)
        {
            Sub1    = r.ReadByte();
            Sub2    = r.ReadByte();
            Value   = new SlimDX.Matrix() {
                M11= r.ReadSingle(), M12= r.ReadSingle(), M13= r.ReadSingle(), M14= r.ReadSingle(),
                M21= r.ReadSingle(), M22= r.ReadSingle(), M23= r.ReadSingle(), M24= r.ReadSingle(),
                M31= r.ReadSingle(), M32= r.ReadSingle(), M33= r.ReadSingle(), M34= r.ReadSingle(),
                M41= r.ReadSingle(), M42= r.ReadSingle(), M43= r.ReadSingle(), M44= r.ReadSingle(),
            };
        }

        public void ReadVector(BinaryReader r)
        {
            Sub1    = r.ReadByte();

            switch(Sub1)
            {
            case 1:     ReadVector3(r); break;
            case 2:     ReadVector4(r); break;
            default:    throw new Exception();
            }
        }

        public void ReadVector3(BinaryReader r)
        {
            Value = new SlimDX.Vector3(r.ReadSingle(), r.ReadSingle(), r.ReadSingle());
        }

        public void ReadVector4(BinaryReader r)
        {
            Value = new SlimDX.Vector4(r.ReadSingle(), r.ReadSingle(), r.ReadSingle(), r.ReadSingle());
        }
        #endregion

        public override string ToString()
        {
            return string.Format("{0:X4}: {1}", ID, Value);
        }
    }
}
