﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SlimDX;

namespace Models.Aqo
{
    using Key = Tuple<Vector3, Vector3>;

    public class TangentCalculator
    {
        public Dictionary<Key, Vector3> Normals                 { get; private set; }
        public Dictionary<Key, Vector3> Tangents                { get; private set; }
        public Dictionary<Key, Vector3> Binormals               { get; private set; }

        public TangentCalculator()
        {
            Normals     = new Dictionary<Key, Vector3>();
            Tangents    = new Dictionary<Key, Vector3>();
            Binormals   = new Dictionary<Key, Vector3>();
        }

        public Vector3 GetNormal(Vector3 pos, Vector3 nrm, Vector3 uv)
        {
            var v   = Vector3.Zero;
          //var key = Tuple.Create(pos, uv);
            var key = Tuple.Create(pos, nrm);

            if(!Normals.TryGetValue(key, out v))
                v   = Vector3.Zero;

            return v;
        }

        public Vector3 GetTangent(Vector3 pos, Vector3 nrm, Vector3 uv)
        {
            var v   = Vector3.Zero;
          //var key = Tuple.Create(pos, uv);
            var key = Tuple.Create(pos, nrm);

            if(!Tangents.TryGetValue(key, out v))
                v   = Vector3.Zero;

            return v;
        }

        public Vector3 GetBinormal(Vector3 pos, Vector3 nrm, Vector3 uv)
        {
            var v   = Vector3.Zero;
          //var key = Tuple.Create(pos, uv);
            var key = Tuple.Create(pos, nrm);

            if(!Binormals.TryGetValue(key, out v))
                v   = Vector3.Zero;

            return v;
        }

        public void AddFace(
            Vector3 p0, Vector3 n0, Vector3 uv0,
            Vector3 p1, Vector3 n1, Vector3 uv1,
            Vector3 p2, Vector3 n2, Vector3 uv2)
        {
            var v0  = Vector3.Normalize(p1-p0);
            var v1  = Vector3.Normalize(p2-p0);
          //var key0= Tuple.Create(p0, uv0);
          //var key1= Tuple.Create(p1, uv1);
          //var key2= Tuple.Create(p2, uv2);
            var key0= Tuple.Create(p0, n0);
            var key1= Tuple.Create(p1, n1);
            var key2= Tuple.Create(p2, n2);

            if(v0.LengthSquared() != 0
            && v1.LengthSquared() != 0)
            {
                var crs = Vector3.Cross(v0, v1);

                if(crs.LengthSquared() != 0)
                {
                    var nrm= Vector3.Normalize(crs);

                    if(nrm.LengthSquared() != 0)
                    {
                        if(Normals  .ContainsKey(key0)) Normals  [key0]+=nrm; else Normals [key0]= nrm;
                        if(Normals  .ContainsKey(key1)) Normals  [key1]+=nrm; else Normals [key1]= nrm;
                        if(Normals  .ContainsKey(key2)) Normals  [key2]+=nrm; else Normals [key2]= nrm;
                    }
                }
            }

            Vector3 tan, bin;

            GeometryUtils.CalcTangentAndBinormal(p0, uv0, p1, uv1, p2, uv2, out tan, out bin);

            if(tan.LengthSquared() != 0
            && bin.LengthSquared() != 0)
            {
                if(Tangents .ContainsKey(key0)) Tangents [key0]+=tan; else Tangents [key0]= tan;
                if(Tangents .ContainsKey(key1)) Tangents [key1]+=tan; else Tangents [key1]= tan;
                if(Tangents .ContainsKey(key2)) Tangents [key2]+=tan; else Tangents [key2]= tan;
                if(Binormals.ContainsKey(key0)) Binormals[key0]+=bin; else Binormals[key0]= bin;
                if(Binormals.ContainsKey(key1)) Binormals[key1]+=bin; else Binormals[key1]= bin;
                if(Binormals.ContainsKey(key2)) Binormals[key2]+=bin; else Binormals[key2]= bin;
            }
        }

        public void Normalize()
        {
            foreach(var i in Normals.Keys.ToArray())
                Normals[i]  = Vector3.Normalize(Normals[i]);

            foreach(var i in Tangents.Keys.ToArray())
                Tangents[i] = Vector3.Normalize(Tangents[i]);

            foreach(var i in Binormals.Keys.ToArray())
                Binormals[i]= Vector3.Normalize(Binormals[i]);
        }
    }
}
