﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Models.Aqo;

namespace Aqp2Obj
{
    class Program
    {
        public static List<string>      Log = new List<string>();

        static int Main(string[] args)
        {
            return Command(args);
        }

        public static int Command(string[] args)
        {
            try
            {
                var files   = new List<string>();
                var config  = Config.FromFile(Config.FileName);

                #if false
                foreach(var i in config.ParamMap)
                    Print("{0}={1}", i.Key, i.Value);
                #endif

                foreach(var i in args)
                {
                    if(i.StartsWith("\""))
                    {
                        files.Add(i.Trim('"'));
                        continue;
                    }

                    var idx = i.IndexOf('=');

                    if(idx < 0)
                    {
                        files.Add(i.Trim('"'));
                    } else
                    {
                        var name    = i.Substring(0, idx);
                        var value   = i.Substring(idx+1);

                        config.Set(name, value);
                    }
                }

                if(files.Count == 0)
                {
                    Usage();
                    return 2;
                }

                foreach(var i in files)
                {
                    switch(Path.GetExtension(i).ToUpper())
                    {
                    case ".OBJ":    ToAqp(i, config);   break;
                    case ".AQP":    ToObj(i, config);   break;
                    case ".AQO":    ToObj(i, config);   break;
                    default:    Console.WriteLine("Unsupported: {0}", i);   break;
                    }
                }

                return 0;
            } catch(Exception ex)
            {
                Print(ex.ToString());

                return 1;
            } finally
            {
                if(Log.Count > 0)
                try
                {
                    var dir     = Path.GetDirectoryName(typeof(Program).Assembly.Location);
                    var logfile = Path.Combine(dir, "lastlog.txt");

                    File.WriteAllLines(logfile, Log);
                } catch {}
            }
        }

        public static void Print(string s)
        {
            Log.Add(s);
            Console.WriteLine(s);
        }

        public static void Print(string fmt, params object[] args)
        {
            Print(string.Format(fmt, args));
        }

        public static void ToAqp(string file, Config config)
        {
            var aqpfile = File.Exists(Path.ChangeExtension(file, ".org.aqp"))
                        ? Path.ChangeExtension(file, ".aqp")
                        : Path.ChangeExtension(file, ".aqo");
            var bakfile = Backup(aqpfile);
            var aqp     = AFPFile.FromFile(bakfile);
            var imp     = new ObjImporter(config);
            var b       = imp.Import(file, aqp);

            Print(string.Format("{0}: {1}", b ? "Complete" : "Failed", file));

            foreach(var i in imp.Messages)
                Print("  {0}", i);
        }

        public static void ToObj(string file, Config config)
        {
            var bakfile = Backup(file);
            var objfile = Path.ChangeExtension(file, ".obj");
            var exp     = new ObjExporter(config);
            var aqp     = AFPFile.FromFile(file);
            var b       = exp.Export(objfile, aqp);

            Print(string.Format("{0}: {1}", b ? "Complete" : "Failed", file));

            foreach(var i in exp.Messages)
                Print("  {0}", i);
        }

        private static void Usage()
        {
            Console.WriteLine("Usage:");
            Console.WriteLine("  Export: .AQP -> .OBJ");
            Console.WriteLine("    aqp2obj <aqofile.aqp>");
            Console.WriteLine("");
            Console.WriteLine("  Import: .OBJ -> .AQO");
            Console.WriteLine("    aqp2obj <objfile.obj>");
        }

        private static string Backup(string file)
        {
            var dir     = Path.GetDirectoryName(file);
            var name    = Path.GetFileNameWithoutExtension(file);
            var ext     = Path.GetExtension(file);
            var bak     = Path.Combine(dir, name+".org"+ext);

            if(!File.Exists(bak))
                File.Copy(file, bak);

            return bak;
        }
    }
}
